#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

int    Filters=0;
int    FastMTF=0;

#ifndef inline
#define inline /* */
#endif

#ifndef __inline
#define __inline inline
#endif

#ifndef __inline__
#define __inline__ __inline
#endif

#ifndef LINUX
#include "include\types.inc"
#include "include\lzp.inc"
#include "include\bwt.inc"
#include "include\st4.inc"
#include "include\gst.inc"
#include "include\block.inc"
#else
#include "include/types.inc"
#include "include/lzp.inc"
#include "include/bwt.inc"
#include "include/st4.inc"
#include "include/gst.inc"
#include "include/block.inc"
#endif

#define CpRMsg          "This is GRZip, yet another block-sorting compressor. Version 0.7.3, 26-Jun-2003.\n"

uc Sign[]=              {71,82,90,105,112,0,7,3};
#define SignSize        8

#define MaxFiles        4096
#define MaxBlockSize    (8*1024*1024-4097)
#define MaxFileNameSize 256

sint    NumFiles=0;
sint    Adaptive=0;
uc  **  Files=NULL;
sint    Stat=0;
sint    BlockSize=MaxBlockSize;
uc  *   ARCName;

void License(void)
{
  fprintf(stdout,"\n   CopyRight (C) 2002-2003 by Grebnov Ilya.\n\n");
  fprintf(stdout,"   This program is free software.\n\n");
  fprintf(stdout,"   This program is distributed in the hope that it will be useful,\n");
  fprintf(stdout,"   but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  fprintf(stdout,"   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.             \n\n");
  fprintf(stdout,"                                             Grebnov Ilya aka GR!B\n\n");
  fprintf(stdout,"                                             E-Mail:grib@crazy.ru\n");
  fprintf(stdout,"                                             http://magics.h10.ru\n");
  fprintf(stdout,"                                             2:5026/49.84@fidonet.org\n");
  fprintf(stdout,"                                             ICQ:965069\n");
  free(Files);
  exit(0);
}


void ShowUsage()
{
  fprintf(stdout,"Usage:GRZip <command> [-<sw> [-<sw>...]] <archive_name> [<file_names>...]\n\n");
  fprintf(stdout,"<Commands>\n");
  fprintf(stdout,"     c     Create new archive\n");
  fprintf(stdout,"     e     Extract files from archive\n");
  fprintf(stdout,"     l     List contents of archive\n\n");
  fprintf(stdout,"<Switches>\n");
  fprintf(stdout,"     -...  Maximum Block size (e.g. -1m, -1024k)\n");
  fprintf(stdout,"           (default -8m or -8192k, maximum -8m)\n");
  fprintf(stdout,"     -f    Use MTF2 instead AWFC and ST order 4 instead BWT\n");
  fprintf(stdout,"     -a    Enable adaptive block size reduction\n");
  fprintf(stdout,"     -i    Enable IE32-x86 filter\n");
  fprintf(stdout,"     -s    Show Statistics\n");
  fprintf(stdout,"     -L    Display software license\n\n");
  fprintf(stdout,"Compression   uses 8*Block_Size     + 1Mb   RAM\n");
  fprintf(stdout,"Decompression uses 4.125*Block_Size + 256Kb RAM\n");
  free(Files);
  exit(0);
}

void Optoins(uc * s)
{
  while (*(++s) != '\0')
    switch (*s)
    {
      case '1': case '2':
      case '3': case '4':
      case '5': case '6':
      case '7': case '8':
      case '9':
        {
          char *ss=s;
          while ((*s >='0')&&(*s <='9')) s++;
          if (*s == 'm')
          {
             BlockSize=atoi(ss)*1024*1024;
             if (BlockSize>MaxBlockSize) BlockSize=MaxBlockSize;
             break;
          }
          if (*s == 'k')
          {
             BlockSize=atoi(ss)*1024;
             if (BlockSize>MaxBlockSize) BlockSize=MaxBlockSize;             
             break;
          }
          ShowUsage();
         }
      case 's':case 'S': { Stat++;break;}
      case 'a':case 'A': { Adaptive=1;break;}
      case 'i':case 'I': { Filters=1;break;}
      case 'f':case 'F': { FastMTF=1;break;}
      case 'l':case 'L': License();
      default : ShowUsage();
   }
}

int CLine(uint argc, char ** argv)
{
  sint Result;
  Files = (uc **) malloc(sizeof(uc * )*MaxFiles);
  argc--;
  fprintf(stderr,CpRMsg);
  if (argc==0) ShowUsage();
  if (*argv[1]=='-') Optoins(argv[1]);
  if (argc<2) ShowUsage();
  if (strlen(argv[1])!=1) ShowUsage();
  switch (*argv[1])
  {
    case 'c':case 'C': { Result=1;argc--; argv++; break; }
    case 'e':case 'E': { Result=2;argc--; argv++; break; }
    case 'l':case 'L': { Result=3;argc--; argv++; break; }
    default : ShowUsage();
  }
  while ((argc>0)&&(*argv[1]=='-'))
  {
    Optoins(argv[1]);
    argc--;
    argv++;
  }
  if (argc==0) ShowUsage();
  ARCName = (uc *) malloc(sizeof(uc)*MaxFileNameSize);
  strcpy(ARCName,argv[1]);
  sint l=strlen(ARCName)-4;
  if ((l<0)||((strcmp(ARCName+l,".grz")!=0)&&(strcmp(ARCName+l,".GRZ")!=0))) strcat(ARCName,".grz");
  argc--;
  argv++;
  while (argc>0)
  {
    Files[NumFiles++]=argv[1];
    argc--;
    argv++;
  }
  return Result;
}

uint FiltTest(uc * Buf,uint Size)
{
  if (Size<48*1024) return 0;
  sint i;
  uint Tab[256];
  uint Win32=0;
  memset(Tab,0,sizeof(Tab));
  for (i=0;i<Size-5;i++)
  {
    uc c=Buf[i];
    Tab[c]++;
    if ((c==0xE8)&&((Buf[i+4]==0)||(Buf[i+4]==0xFF))) Win32++;
    if ((c==0xE9)&&((Buf[i+4]==0)||(Buf[i+4]==0xFF))) Win32++;
    if ((c==0x0F)&&((Buf[i+1]&0xF0)==0x80)&&((Buf[i+5]==0)||(Buf[i+5]==0xFF))) Win32++;
  }
  if (Win32*200>Size) return 1;
  return 0;
}

void UnPackBufFilt(BlockSign * BS,FILE * InFile,FILE * OutFile)
{
  if (BS->Flag==0)
  {
    uc * Buf=malloc(BS->Size);
    fread(Buf,1,BS->Size,InFile);
    uc * OutPut = UnPackBuf(Buf,BS);
    fwrite(OutPut,1,BS->Data,OutFile);
    free(OutPut);
    return;
  }
  if (BS->Flag==1)
  {
    UnPackBufWin32(BS,InFile,OutFile);
    return;    
  }
  
}

void PackBufFilt(uc * * Buf,uint Size,FILE *Arc)
{
  BlockSign BS;
  memset(&BS,0,sizeof(BS));
  uint Tst=(Filters)&&(FiltTest(* Buf,Size));
  if (Tst==0)
  {
    uc * OutPut = PackBuf(Buf,Size,&BS);
    fwrite(&BS,1,sizeof(BS),Arc);
    fwrite(OutPut,1,BS.Size,Arc);
    free(OutPut);
    return;
  }
  if (Tst==1)
  {
    PackBufWin32(Buf,Size,Arc);
    return;
  }  
}

uint AddFile(FILE * Arc,uc * FileName)
{
  FILE * f =fopen(FileName,"rb");
  if (f==NULL)
  {
    fprintf(stdout,"Error! File not found!\n");
    return 0;
  }
  
  uint FBeg=ftell(Arc);
  FileSign FS;

  memset(&FS,0,sizeof(FileSign));
  
  fseek(f,0,SEEK_END);
  FS.RealSize=ftell(f);
  fseek(f,0,SEEK_SET);
  
  FS.FNameSize=strlen(FileName)+1;

  fwrite(&FS,1,sizeof(FileSign),Arc);
  fwrite(FileName,1,FS.FNameSize,Arc);

  while ((feof(f)==0)&&(ftell(f)!=FS.RealSize))
  {
    uc * InPut=malloc(BlockSize);
    uint NumRead=fread(InPut,1,BlockSize,f);
    FS.NumSect++;
    if (Adaptive)
    {
       sint nSize=AdaptiveBlock(InPut,NumRead);
       sint NewSec=ftell(f)+nSize-NumRead;
       fseek(f,NewSec,SEEK_SET);
       NumRead=nSize;
    }
    PackBufFilt(&InPut,NumRead,Arc);
    fprintf(stdout,"\r                                                                       \r");
    fflush(stdout);    
    fprintf(stdout,"Adding %.55s(%02d%%)",FileName,(100*ftell(f))/FS.RealSize);
    fflush(stdout);
    free(InPut);
  }

  FS.Next=ftell(Arc)-sizeof(FileSign)-FS.FNameSize-FBeg;
  fseek(Arc,FBeg,SEEK_SET);
  fwrite(&FS,1,sizeof(FileSign),Arc);  
  fseek(Arc,0,SEEK_END);
  fclose(f);
  return FS.RealSize;
}

void Create(void)
{
  if (NumFiles==0) ShowUsage();
  FILE * Arc=fopen(ARCName,"wb+");
  if (Arc==NULL)
  {
    fprintf(stderr,"Can'n create new archive!\n");
    return;
  }
  fwrite(Sign,1,SignSize,Arc);
  uint Nf, TotSize=0;
  clock_t BeginTime = clock();
  for (Nf=0;Nf<NumFiles;Nf++)
  {
    fprintf(stdout,"Adding %.55s",Files[Nf]);fflush(stdout);
    fflush(stdout);    
    TotSize += AddFile(Arc,Files[Nf]);
    fprintf(stdout,"\r                                                                       \r");
    fflush(stdout);
  }
  if (Stat) CreateStat(clock()-BeginTime);
  uint ArcSize=ftell(Arc);
  float TotTime=((float)(clock()-BeginTime))/CLOCKS_PER_SEC;
  fprintf(stdout,"%d files added! Compressed %d into %d in %.3f seconds.\n"
  ,Nf,TotSize,ArcSize,TotTime);
  fclose(Arc);
}

void Extract(void)
{
  FILE * Arc=fopen(ARCName,"rb");
  if (Arc==NULL)
  {
    fprintf(stderr,"Can'n open archive!\n");
    return;
  }
  uc SignTest[SignSize];
  if (fread(SignTest,1,SignSize,Arc)!=SignSize)
  {
    fclose(Arc);
    fprintf(stderr,"This's not GRZip archive!\n");
    return;
  }
  if (memcmp(SignTest,Sign,SignSize)!=00)
  {
    fclose(Arc);
    fprintf(stderr,"This's not GRZip archive!\n");
    return;
  }
  fseek(Arc,0,SEEK_END);
  uint ArcSize=ftell(Arc);
  fseek(Arc,SignSize,SEEK_SET);
  clock_t BeginTime = clock();
  sint Nf=0;
  uint TotSize=0;
  while ((feof(Arc)==0)&&(ftell(Arc)!=ArcSize))
  {
    FileSign FS;
    Nf++;
    fread(&FS,1,sizeof(FileSign),Arc);
    uc * FileName=malloc(FS.FNameSize);
    fread(FileName,1,FS.FNameSize,Arc);
    FILE * f=fopen(FileName,"wb+");
    uint BC;
    fprintf(stdout,"Extracting %.55s\r",FileName);fflush(stdout);
    for (BC=0;BC<FS.NumSect;BC++)
    {
      BlockSign BS;
      fread(&BS,1,sizeof(BlockSign),Arc);
      uint k=ftell(f);
      UnPackBufFilt(&BS,Arc,f);
      fprintf(stdout,"\r                                                                     \r");
      fflush(stdout);
      fprintf(stdout,"Extracting %.55s(%02d%%)",FileName,(100*(BC+1))/FS.NumSect);fflush(stdout);
      TotSize += (ftell(f)-k);
    }
      fprintf(stdout,"\r                                                                     \r");
      fflush(stdout);
    fclose(f);
    free(FileName);
  }  
  if (Stat) ExtractStat(clock()-BeginTime);
  float TotTime=((float)(clock()-BeginTime))/CLOCKS_PER_SEC;
  fprintf(stdout,"%d files extracted! Decompressed %d into %d in %.3f seconds.\n"
  ,Nf,ArcSize,TotSize,TotTime);
  fclose(Arc);  
}

void ListArc(void)
{
  FILE * Arc=fopen(ARCName,"rb");
  if (Arc==NULL)
  {
    fprintf(stderr,"Can'n open archive!\n");
    return;
  }
  uc SignTest[SignSize];
  if (fread(SignTest,1,SignSize,Arc)!=SignSize)
  {
    fclose(Arc);
    fprintf(stderr,"This's not GRZip archive!\n");
    return;    
  }
  if (memcmp(SignTest,Sign,SignSize)!=00)
  {
    fclose(Arc);
    fprintf(stderr,"This's not GRZip archive!\n");
    return;    
  }
  fseek(Arc,0,SEEK_END);
  uint ArcSize=ftell(Arc);
  fseek(Arc,SignSize,SEEK_SET);
  fprintf(stdout,"     Size     CompSize   Sections       bpc            Name\n");
  fprintf(stdout,"   --------   --------   --------   -----------    --------------\n");
  uint TotSize=0,NumFiles=0;
  uint bpc=0;
  double Avgbpc=0;
  sint count,TotSect=0;
  while ((feof(Arc)==0)&&(ftell(Arc)!=ArcSize))
  {
    FileSign FS;
    fread(&FS,1,sizeof(FileSign),Arc);
    uc * FileName=malloc(FS.FNameSize);
    fread(FileName,1,FS.FNameSize,Arc);
    fprintf(stdout,"%11u %10u      %02u         %0.3f       %-20s\n",
    FS.RealSize,FS.Next,FS.NumSect,((double)(8*FS.Next))/FS.RealSize,FileName);
    TotSect += FS.NumSect;
    bpc += FS.Next;
    Avgbpc += ((double)(8*FS.Next))/FS.RealSize;
    count++;
    TotSize +=FS.RealSize;NumFiles++;
    fseek(Arc,FS.Next,SEEK_CUR);
    free(FileName);
  }
  Avgbpc /= NumFiles;
  fprintf(stdout,"   --------   --------   --------   -----------    --------------\n");
    fprintf(stdout,"%11u %10u      %02u      %0.3f|%0.3f    %2d files\n",
    TotSize,bpc,TotSect,((double)(8*bpc))/TotSize,Avgbpc,NumFiles);
//  fprintf(stdout,"%11u %10u           %2d files\n",TotSize,bpc,NumFiles);
  fclose(Arc);
}

int main(int argc,char ** argv)
{
  switch (CLine(argc, argv))
  {
    case 1: {Create();break;}
    case 2: {Extract();break;}
    case 3: {ListArc();break;}
  }
  free(Files);
  return 0;
}

