uint PackTotSize=0;
uint LZPTimes=0;
uint BWTTimes=0;
uint AWFCTimes=0;
uint ARITimes=0;

uc * PackBuf(uc * * InPut,uint NumRead,BlockSign * BS)
{
  PackTotSize += NumRead;
  BS->Data=NumRead;
  if (NumRead<32)
  {
    BS->FBP=0xFFFFFFFF;
    uc * OutPut=malloc(NumRead);
    BS->Size=NumRead;
    memcpy(OutPut,* InPut,NumRead);
    return OutPut;
  }
  uc * Tmp=(uc *)malloc(2*NumRead);
  clock_t T0=clock();
  uint Size=LZ_Preprocess(* InPut,Tmp,NumRead);
  LZPTimes += clock()-T0;T0=clock();
  if (!FastMTF)
    BS->FBP=BWTBlock(Tmp,Size);
  else
    BS->FBP=GRZip_ST4_Encode(Tmp,Size);
  BWTTimes += clock()-T0;T0=clock();
  uc * OutPut=malloc(Size+1024);
  if (!FastMTF)
    GRZip_AWFC_Encode(Tmp,Size);
  else
  {
    GRZip_MTF_Encode(Tmp,Size);
    BS->FBP|=0x40000000;
  }
  AWFCTimes += clock()-T0;T0=clock();
  BS->Size=EnCodeMTF(Tmp,Size,OutPut);
  ARITimes += clock()-T0;
  if (BS->Size==0)
  {
    BS->FBP=0xFFFFFFFF;
    BS->Size=NumRead;
    memcpy(OutPut,* InPut,NumRead);    
  }
  free(Tmp);
  return OutPut;
}

uc * UnPackBuf(uc * Input, BlockSign * BS)
{
  uc * OutPut=malloc(BS->Data);
  if (BS->FBP==0xFFFFFFFF)
  {
    memcpy(OutPut, Input,BS->Data);
    return OutPut;
  }
  uc * Tmp=(uc *)malloc(2*BS->Data);  
  clock_t T0=clock();
  uint Size=DeCodeMTF(Input,Tmp);
  ARITimes += clock()-T0;T0=clock();
  free(Input);
  if ((BS->FBP&0x40000000)==0)
  {
    GRZip_AWFC_Decode(Tmp,Size);
    AWFCTimes += clock()-T0;T0=clock();
    UnBWTBlock(Tmp,Size,BS->FBP);
    BWTTimes += clock()-T0;
  }
  else
   {
     FastMTF=1;
     GRZip_MTF_Decode(Tmp,Size);
     BS->FBP&=0x3FFFFFFF;
     AWFCTimes += clock()-T0;T0=clock();
     GRZip_ST4_Decode(Tmp,Size,BS->FBP);
     BWTTimes += clock()-T0;
   }
  T0=clock();
  LZ_UnPreprocess(Tmp,OutPut,Size);
  LZPTimes += clock()-T0;T0=clock();
  free(Tmp);
  return OutPut;
}

#define MaxAdpBlockSize 16*1024

#define lg log10

sint AdaptiveBlock(uc * Buf,sint BlockSize)
{

  if (BlockSize<MaxAdpBlockSize) return BlockSize;
  uint FullTab[256];
  uint BlockTab[256];
  uint Freq[256];
  memset(FullTab,0,256*sizeof(uint));
  sint Pos, AdpBlockSize=MaxAdpBlockSize / 2;
  for (Pos=0;Pos<AdpBlockSize;Pos++) FullTab[Buf[Pos]]++;
  
  while (Pos+AdpBlockSize<BlockSize)
  {
    memset(BlockTab,0,256*sizeof(uint));
    sint i=0,Sum=0;
    float AvgSize=0,RealSize=0;
    for (;i<AdpBlockSize;i++) {BlockTab[Buf[i+Pos]]++;};
    for (i=0;i<256;i++)
      if (BlockTab[i])
        RealSize -= BlockTab[i]*lg((float)BlockTab[i]/AdpBlockSize);

    for (i=0;i<256;i++)
      if (BlockTab[i])
      {
        Freq[i] = BlockTab[i] + (FullTab[i] >> 1);
        Sum += Freq[i];
      }

    for (i=0;i<256;i++)
     if (BlockTab[i])
       AvgSize -= BlockTab[i]*lg((float)Freq[i]/Sum);

    if (AvgSize>1.25*RealSize)
       if (AdpBlockSize<1024)
         return Pos;
       else
       {
         AdpBlockSize >>= 1;
         continue;
       }
    for (i=0;i<256;i++)
      FullTab[i] += BlockTab[i];
    Pos += AdpBlockSize;
  }

  return BlockSize;
}

void CreateStat(uint TotTime)
{
  uint MiscTimes=TotTime-BWTTimes-ARITimes-AWFCTimes-LZPTimes;
  if (TotTime)
  {
    float T=((float)(LZPTimes))/CLOCKS_PER_SEC;
    double f=((float)(100*LZPTimes))/TotTime;
      fprintf(stdout,"LZP Preprocess      = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(BWTTimes))/CLOCKS_PER_SEC;
    f=((float)(100*BWTTimes))/TotTime;
    if (!FastMTF)
      fprintf(stdout,"BW Transform        = %06.03f (%05.02f%%) seconds.\n",T,f);
    else
      fprintf(stdout,"Schindler Transform = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(AWFCTimes))/CLOCKS_PER_SEC;
    f=((float)(100*AWFCTimes))/TotTime;
    if (!FastMTF)
      fprintf(stdout,"AWFC Encoding       = %06.03f (%05.02f%%) seconds.\n",T,f);
    else
      fprintf(stdout,"MTF2 Encoding       = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(ARITimes))/CLOCKS_PER_SEC;
    f=((float)(100*ARITimes))/TotTime;
    fprintf(stdout,"Modeling & Encoding = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(MiscTimes))/CLOCKS_PER_SEC;
    f=((float)(100*MiscTimes))/TotTime;
    fprintf(stdout,"Miscellaneous       = %06.03f (%05.02f%%) seconds.\n\n",T,f);
  }
}

void ExtractStat(uint TotTime)
{
  uint MiscTimes=TotTime-BWTTimes-ARITimes-AWFCTimes-LZPTimes;
  if (TotTime)
  {
    float T=((float)(LZPTimes))/CLOCKS_PER_SEC;
    double f=((float)(100*LZPTimes))/TotTime;
    fprintf(stdout,"Reverse LZP         = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(BWTTimes))/CLOCKS_PER_SEC;
    f=((float)(100*BWTTimes))/TotTime;
    if (!FastMTF)
      fprintf(stdout,"Reverse BWT         = %06.03f (%05.02f%%) seconds.\n",T,f);
    else
      fprintf(stdout,"Reverse ST          = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(AWFCTimes))/CLOCKS_PER_SEC;
    f=((float)(100*AWFCTimes))/TotTime;
    if (!FastMTF)
      fprintf(stdout,"AWFC Decoding       = %06.03f (%05.02f%%) seconds.\n",T,f);
    else
      fprintf(stdout,"MTF2 Decoding       = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(ARITimes))/CLOCKS_PER_SEC;
    f=((float)(100*ARITimes))/TotTime;
    fprintf(stdout,"Modeling & Decoding = %06.03f (%05.02f%%) seconds.\n",T,f);
    T=((float)(MiscTimes))/CLOCKS_PER_SEC;
    f=((float)(100*MiscTimes))/TotTime;
    fprintf(stdout,"Miscellaneous       = %06.03f (%05.02f%%) seconds.\n\n",T,f);
  }
}

#define JTabSize 255
#define CTabSize 255

uint JTab[JTabSize+1];
uint CTab[CTabSize+1];


void UnPackBufWin32(BlockSign * BS,FILE * InFile,FILE * OutFile)
{
  sint   Size     = BS->Data;
  uc * OutPut     = malloc(Size);

  Filters=0;

  fread(BS,1,sizeof(BlockSign),InFile);
  uc * Buf=malloc(BS->Size);
  fread(Buf,1,BS->Size,InFile);
  uc * Code = UnPackBuf(Buf,BS);

  fread(BS,1,sizeof(BlockSign),InFile);
  Buf=malloc(BS->Size);
  fread(Buf,1,BS->Size,InFile);
  uc * JCode0 = UnPackBuf(Buf,BS);

  fread(BS,1,sizeof(BlockSign),InFile);
  Buf=malloc(BS->Size);
  fread(Buf,1,BS->Size,InFile);
  uint * JCode1 =(uint *)UnPackBuf(Buf,BS);

  fread(BS,1,sizeof(BlockSign),InFile);
  Buf=malloc(BS->Size);
  fread(Buf,1,BS->Size,InFile);
  uc * CCode0 = UnPackBuf(Buf,BS);

  fread(BS,1,sizeof(BlockSign),InFile);
  Buf=malloc(BS->Size);
  fread(Buf,1,BS->Size,InFile);
  uint * CCode1 =(uint *)UnPackBuf(Buf,BS);

  memset(JTab,0,(JTabSize+1)*sizeof(uint));
  memset(CTab,0,(CTabSize+1)*sizeof(uint));

  sint i,CodeP=0,C0CodeP=0,C1CodeP=0,J0CodeP=0,J1CodeP=0;
  for (i=0;i<Size;i++)
  {
    uc c=Code[CodeP++];
    OutPut[i]=c;
    if ((c==0xE8)&&(i<Size-5))
    {
      uint call;
      uc pos=CCode0[C0CodeP++];
      if (pos==CTabSize)
         call=CCode1[C1CodeP++];
      else
         call=CTab[pos];
      while (pos>0)
      {
        CTab[pos]=CTab[pos-1];
        pos--;
      }
      CTab[0]=call;
      call -= i;
      (* (uint*) (OutPut+i+1))=call;
      i += 4;
    }
    if ((c==0xE9)&&(i<Size-5))
    {
      uint jmp;
      uc pos=JCode0[J0CodeP++];
      if (pos==JTabSize)
         jmp=JCode1[J1CodeP++];
      else
         jmp=JTab[pos];
      while (pos>0)
      {
        JTab[pos]=JTab[pos-1];
        pos--;
      }
      JTab[0]=jmp;
      jmp -= i;
      (* (uint*) (OutPut+i+1))=jmp;
      i += 4;
    }    
  }
  fwrite(OutPut,1,Size,OutFile);

  Filters=1;
  
  free(CCode1);
  free(CCode0);
  free(JCode1);  
  free(JCode0);
  free(Code);
  free(OutPut);
}

extern void PackBufFilt(uc * * Buf,uint Size,FILE *Arc);

void PackBufWin32(uc * * InPut,uint Size,FILE *Arc)
{
  memset(JTab,0,(JTabSize+1)*sizeof(uint));
  memset(CTab,0,(CTabSize+1)*sizeof(uint));
  uc   * Buf     = * InPut;
  uc   * Code    = malloc(Size);
  uc   * JCode0  = malloc(Size);
  uint * JCode1  = malloc(Size);
  uc   * CCode0  = malloc(Size);
  uint * CCode1  = malloc(Size);
  sint CodeSize=0,J0Size=0,J1Size=0,C0Size=0,C1Size=0;
  sint i;
  
  for (i=0;i<Size;i++)
  {
    uc c=Buf[i];Code[CodeSize++]=c;

    if ((c==0xE8)&&(i<Size-5))
    {
       uint call=i+ (* (uint*) (Buf+i+1));
       CTab[CTabSize]=call;
       sint pos=0;
       while (CTab[pos]!=call) pos++;
       CCode0[C0Size++]=pos;
       if (pos==CTabSize) CCode1[C1Size++]=call;
       while (pos>0)
       {
         CTab[pos]=CTab[pos-1];
         pos--;
       }
       CTab[0]=call;
       i += 4;
    }
    if ((c==0xE9)&&(i<Size-5))
    {
       uint jmp=i+(* (uint *) (Buf+i+1));
       JTab[JTabSize]=jmp;
       sint pos=0;
       while (JTab[pos]!=jmp) pos++;
       JCode0[J0Size++]=pos;
       if (pos==JTabSize) JCode1[J1Size++]=jmp;
       while (pos>0)
       {
         JTab[pos]=JTab[pos-1];
         pos--;
       }
       JTab[0]=jmp;
       i += 4;
    }
  }

  BlockSign BS;
  memset(&BS,0,sizeof(BlockSign));
  BS.Flag=1;
  BS.Data=Size;
  fwrite(&BS,1,sizeof(BlockSign),Arc);

  Filters=0;
  PackBufFilt(&Code,CodeSize,Arc);
  PackBufFilt(&JCode0,J0Size,Arc);
  uc * Tmp=(uc *)(JCode1);
  PackBufFilt(&Tmp,4*J1Size,Arc);
  PackBufFilt(&CCode0,C0Size,Arc);
       Tmp=(uc *)(CCode1);
  PackBufFilt(&Tmp,4*C1Size,Arc);

  Filters=1;

  free(Code);
  free(JCode0);
  free(JCode1);
  free(CCode0);
  free(CCode1);
  return;
}

