#define BWT_MinQSort          20
#define BWT_MedTrh            64
#define BWT_MaxByte           256
#define BWT_MaxWord           65536
#define BWT_TSortStackSize    3*65536

#define BWT_BFreq(B)          (Buckets[(B+1) << 8] - Buckets[B << 8])
#define BWT_SPush(L,H,D)      {Stack[sp]=L;Stack[sp+1]=H;Stack[sp+2]=D;sp+=3;}
#define BWT_SPop(L,H,D)       {sp-=3;L=Stack[sp];H=Stack[sp+1];D=Stack[sp+2];}
#define BWT_Swap(SC1,SC2)     {sint Stmp=SC1;SC1=SC2;SC2=Stmp;}
#define BWT_Min(A,B)          ((A) < (B)) ? (A): (B)
#define BWT_CMask             0x3FFFFFFF
#define BWT_SMask             0x40000000

#define BWT_VSwap(S1,S2,Num)             \
{                                        \
  sint TS1=(S1);                         \
  sint TS2=(S2);                         \
  sint TNum=(Num);                       \
  while (TNum)                           \
  {                                      \
    BWT_Swap(Index[TS1],Index[TS2]);     \
    TS1++;TS2++;TNum--;                  \
  }                                      \
}

__inline__ static
uint BWT_Med3(uint MA,uint MB,uint MC)
{
  uint MT;
  if (MA>MB) {MT=MA;MA=MB;MB=MT;}
  if (MB>MC)
  {
    MB=MC;
    if (MA>MB) MB=MA;
  }
  return MB;
}

__inline__ static
uc SimpleCMP(uint * Group,sint i1,sint i2)
{
  uint CMP1,CMP2;
  CMP1=Group[i1];CMP2=Group[i2];
  while (CMP1==CMP2) {CMP1=Group[i1+=2];CMP2=Group[i2+=2];}
  return (CMP1>CMP2);
}

__inline__ static
void ShellSort(sint * Index, uint * Group,sint Lo,sint Hi,sint d)
{
  sint i,j,s,h;
  sint n=Hi-Lo+1;
  if (n<2)
  {
    Group[Index[Lo]]=(Group[Index[Lo]]&0xFF000000)|Lo;
    return;
  }
  h=1;
  while (h<=n) h=h*3+1;
  do {
    h /= 3;
    for (i=Lo+h;i<=Hi;i++)
    {
      s=Index[i]+d;
      j=i;
      while (SimpleCMP(Group,Index[j-h]+d,s))
      {
        Index[j]=Index[j-h];
        j -= h;
        if (j<h+Lo) break;
      }
      Index[j]=s-d;
    }
  } while (h!=1);
  for (;Lo<=Hi;Lo++)
    Group[Index[Lo]]=(Group[Index[Lo]]&0xFF000000)|Lo;
}

static
void ShellSortDeph2(sint * Index, uint * Group,sint Lo,sint Hi)
{
  sint i,j,s;
  if (Hi==Lo)
  {
    Group[Index[Lo]]=(Group[Index[Lo]]&0xFF000000)|Lo;
    return;
  }
  if (Hi-Lo>7)
    for (i=Lo+4;i<=Hi;i++)
    {
      s=Index[i]+2;
      j=i-4;
      while (SimpleCMP(Group,Index[j]+2,s))
      {
        Index[j+4]=Index[j];
        if ((j-=4)<Lo) break;
      }
      Index[j+4]=s-2;
    }
  for (i=Lo+1;i<=Hi;i++)
  {
    s=Index[i]+2;
    j=i-1;
    while (SimpleCMP(Group,Index[j]+2,s))
    {
      Index[j+1]=Index[j];
      if ((--j)<Lo) break;
    }
    Index[j+1]=s-2;
  }
  for (;Lo<=Hi;Lo++)
    Group[Index[Lo]]=(Group[Index[Lo]]&0xFF000000)|Lo;
}

static
void TernarySort(sint * Index,uint * Group,sint Lo,sint Hi,sint * Stack)
{
  sint A,B,C,D,N,M,d,sp=0;
  BWT_SPush(Lo,Hi,2);
  while (sp)
  {
    BWT_SPop(Lo,Hi,d);
    if (Hi<Lo) continue;

    if ((Hi-Lo<BWT_MinQSort)||(sp>=BWT_TSortStackSize))
    {
      ShellSort(Index,Group,Lo,Hi,d);
      continue;
    }
    uint Med;
    if ((Hi-Lo)<BWT_MedTrh)
      Med=BWT_Med3(Group[Index[(Lo+Hi)>>1]+d],Group[Index[Lo]+d],Group[Index[Hi]+d]);
    else
    {
      M=(Hi-Lo)>>1;
      N=(Hi-Lo)>>3;
      uint Med1=BWT_Med3(Group[Index[Lo]+d],Group[Index[Lo+N]+d],Group[Index[Lo+N+N]+d]);
      uint Med2=BWT_Med3(Group[Index[M-N]+d],Group[Index[M]+d],Group[Index[M+N]+d]);
      uint Med3=BWT_Med3(Group[Index[Hi-N-N]+d],Group[Index[Hi-N]+d],Group[Index[Hi]+d]);
      Med=BWT_Med3(Med1,Med2,Med3);
    }
    A=Lo;B=Lo;C=Hi;D=Hi;
    while (1)
    {
      while (1)
      {
        if (B>C) break;
        if (Group[Index[B]+d]==Med)
        {
          BWT_Swap(Index[A],Index[B]);
          A++;B++;
          continue;
        }
        if (Group[Index[B]+d]>Med) break;
        B++;
      }
      while (1)
      {
        if (B>C) break;
        if (Group[Index[C]+d]==Med)
        {
          BWT_Swap(Index[C],Index[D]);
          D--;C--;
          continue;
        }
        if (Group[Index[C]+d]<Med) break;
        C--;
      }      
    if (B>C) break;
    BWT_Swap(Index[B],Index[C]);
    B++;C--;
  }

  if (D<A)
  {
    BWT_SPush(Lo,Hi,d+2);
    continue;
  }
  
  N=BWT_Min(A-Lo,B-A);BWT_VSwap(Lo,B-N,N);
  M=BWT_Min(Hi-D,D-C);BWT_VSwap(B,Hi-M+1,M);

  N=Lo+B-A-1;
  M=Hi-(D-C)+1;

  BWT_SPush(M,Hi,d);
  BWT_SPush(N+1,M-1,d+2);
  BWT_SPush(Lo,N,d);
  }
}  

sint BWTBlock(uc * Input,sint Size)
{
  uc      RunOrder[BWT_MaxByte],BigDone[BWT_MaxByte];
  sint    CopyStart[BWT_MaxByte],CopyEnd[BWT_MaxByte];
  uint *  Group  =(uint *)malloc((Size+1)*sizeof(uint));
  sint *  Index  =(sint *)malloc((Size+1)*sizeof(sint));
  sint *  Buckets=(sint *)malloc((BWT_MaxWord+1)*sizeof(sint));
  sint *  BGroups=(sint *)malloc((BWT_TSortStackSize+3)*sizeof(sint));
  memset  (Buckets,0,(BWT_MaxWord+1)*sizeof(sint));
  sint i;
  uc   LB=Input[Size-1];
  uint W=LB<<8;
  for (i=Size-2;i>=0;i--) Buckets[(W=(W>>8)|(Input[i]<<8))]++;
  Buckets[LB<<8]++;
  BGroups[0]=Buckets[0];
  for (i=1;i<=BWT_MaxWord;i++) BGroups[i]=(Buckets[i]+=Buckets[i-1]);
  W=LB<<8;
  for (i=Size-2;i>=0;i--)
  {
    W=(W>>8)|(Input[i]<<8);
    Group[i]=BGroups[W]|(Input[i]<<24);
    Index[Buckets[W]--]=i;
  }
  Group[Size]=0;Group[Size-1]=Buckets[LB<<8]|(LB<<24);
  Index[0]=Size;Index[Buckets[LB<<8]--]=Size-1;
  for (i=0;i<BWT_MaxByte;i++){RunOrder[i]=i;BigDone[i]=0;}
  sint h=364;
  do {
    h /= 3;
    for (i=h;i<BWT_MaxByte;i++)
    {
      uc s=RunOrder[i];
      sint j=i;
      while (BWT_BFreq(RunOrder[j-h])>BWT_BFreq(s))
      {
        RunOrder[j]=RunOrder[j-h];
        j -= h;
        if (j<h) break;
      }
      RunOrder[j]=s;
    }
  } while (h!=1);
  h=0; while (BWT_BFreq(RunOrder[h])==0) h++;
  while (h<BWT_MaxByte)
  {
    uc S=RunOrder[h];
    sint j;
    for (j=0;j<BWT_MaxByte;j++)
      if (j!=S)
      {
        sint k=(S<<8)|j;
        sint Lo=(Buckets[k])+1;
        sint Hi=(Buckets[k+1]&BWT_CMask);
        if ((LB<<8)==k) Lo++;
        if ((Lo<BWT_SMask)&&(Hi>=Lo))
          if (Hi-Lo>BWT_MinQSort)
            TernarySort(Index,Group,Lo,Hi,BGroups);
          else
            ShellSortDeph2(Index,Group,Lo,Hi);
      }
    for (j=0;j<BWT_MaxByte;j++)
    {
      CopyStart[j]=(Buckets[(j<<8)+S]&BWT_CMask)+1;
      if ((LB<<8)==((j<<8)+S)) CopyStart[j]++;
      Buckets[(j<<8)+S] |= BWT_SMask;
      CopyEnd[j]=(Buckets[(j<<8)+S+1] & BWT_CMask);
    }
    for (j=(Buckets[S<<8]&BWT_CMask)+1;j<CopyStart[S];j++)
    {
      sint k=Index[j]-1;
      if (k<0) continue;
      uc c=Group[k]>>24;
      if (!BigDone[c])
      {
        Group[k]=(Group[k]&0xFF000000)|CopyStart[c];
        Index[CopyStart[c]++]=k;
      }
    }
    for (j=(Buckets[(S+1)<<8]&BWT_CMask);j>CopyEnd[S];j--)
    {
      sint k=Index[j]-1;
      if (k<0) continue;
      uc c=Group[k]>>24;
      if (!BigDone[c])
      {
        Group[k]=(Group[k]&0xFF000000)|CopyEnd[c];
        Index[CopyEnd[c]--]=k;
      }
    }
    BigDone[S]=1;
    h++;
  }
  free(Buckets);
  free(BGroups);
  uc Pred=Input[0];
  W= Group[0]&0xFFFFFF;
  for (i=1;i<=Size;i++)
  {
    uint Ps=(Group[i]&0xFFFFFF);
    uc   c=Group[i]>>24;
    if (Ps<W)
      Input[Ps]=Pred;
    else
      Input[Ps-1]=Pred;
    Pred=c;
  }
  free(Index);
  free(Group);
  return W;
}

void UnBWTBlock(uc * Input,sint Size,sint FBP)
{
  uint * T=(uint *)malloc((Size+1)*sizeof(uint));
  sint Count[BWT_MaxByte];
  sint i,Sum;
  memset(Count,0,BWT_MaxByte*sizeof(uint));
  for (i=0;i<FBP;i++)
  {
    uc c=Input[i]; T[i]=((Count[c]++)<<8)|c;
  }
  for (i=FBP;i<Size;i++)
  {
    uc c=Input[i]; T[i+1]=((Count[c]++)<<8)|c;
  }
  for (Sum=1,i=0;i<BWT_MaxByte;i++){Sum+=Count[i];Count[i]=Sum-Count[i];}
  for (FBP=0,i=Size-1;i>=0;i--)
  {
    uint u=T[FBP];
    uc   c=u&0xFF;
    FBP=(u>>8)+Count[c];
    Input[i]=c;
  }
  free(T);
}

#undef BWT_MinQSort
#undef BWT_CMask
#undef BWT_SMask
#undef BWT_MaxByte
#undef BWT_MaxWord
#undef BWT_TSortStackSize
#undef BWT_BFreq
#undef BWT_SPush
#undef BWT_SPop
#undef BWT_Swap
#undef BWT_Min
#undef BWT_VSwap
