
uc **HashTable4, **HashTable5;

#define LowerLimit 45
#define HTSIZE4 65536UL
#define HTSIZE5 32768UL

__inline__ static
uint HashFunction4(uint index,uc *PTR)
{
  uint x;
  x=(((uint)PTR[index-4])<<24) | (((uint)PTR[index-3])<<16) |
    (((uint)PTR[index-2])<<8) | (((uint)PTR[index-1]));
  x=(x>>15)^x^(x>>3);
  return x&(HTSIZE4-1);
}

__inline__ static
uint HashFunction5(uint index,uc *PTR)
{
  uint x;
  x=(((uint)PTR[index-4])<<24) | (((uint)PTR[index-3])<<16) |
    (((uint)PTR[index-2]))<<8 | (((uint)PTR[index-1]));
  x=(x>>25 | ((uint)PTR[index-5])<<7)^x^(x<<4);
  return x&(HTSIZE5-1);
}

__inline__ static
void OutPutLength(uint OutPutLength, uc *OutBuffer, uint *PBuffer)
{
 while (OutPutLength>254) {
       OutBuffer[(*PBuffer)++]=255;
       OutPutLength-=255;
 }
 OutBuffer[(*PBuffer)++]=(uc)OutPutLength;
}

__inline__ static
uint GetLength(uc *InputBuffer, uint *PBuff)
{
  uint Result;
  Result=0;
  while (Result+=(uint)InputBuffer[*PBuff],InputBuffer[(*PBuff)++]==255);
  return Result;
}

void CleanTabs(void)
{
 uint i;
 for (i=0;i<HTSIZE4;i++) HashTable4[i]=NULL;
 for (i=0;i<HTSIZE5;i++) HashTable5[i]=NULL;
}

uint LZ_Preprocess(uc *InData, uc *OutData, uint InLength)
{
 uint Pointer,i,CommonLength,OutLength,HashIndex4,HashIndex5;
 uc *Pointer4, *Pointer5, *PastPointer;

 HashTable4=(uc **)malloc(HTSIZE4*(uint)sizeof(uc *));
 HashTable5=(uc **)malloc(HTSIZE5*(uint)sizeof(uc *));

 CleanTabs();

 OutLength=0;
 for (i=0; i<5; i++) OutData[i]=InData[i];
 OutLength+=5;

 Pointer=5;

 HashTable4[HashFunction4(4,InData)]=InData+4;
 while (Pointer<InLength) {

       HashIndex4=HashFunction4(Pointer,InData);
       HashIndex5=HashFunction5(Pointer,InData);

       Pointer4=HashTable4[HashIndex4];
       Pointer5=HashTable5[HashIndex5];

       HashTable5[HashIndex5]=HashTable4[HashIndex4]=InData+Pointer;

       if (Pointer5!=NULL || Pointer4!=NULL) {
          if (Pointer5!=NULL) PastPointer=Pointer5;
           else PastPointer=Pointer4;

          CommonLength=0;

          while (Pointer<InLength) {
                if (InData[Pointer]!=*PastPointer) break;
                Pointer++;
                PastPointer++;
                CommonLength++;
          }
          if (CommonLength>0 && CommonLength<LowerLimit) {
             Pointer-=CommonLength;
             CommonLength=0;
          }
          if (CommonLength)
             OutPutLength(CommonLength-LowerLimit+256UL, OutData, &OutLength);
           else
             OutPutLength((uint)InData[Pointer++], OutData, &OutLength);
       } else  OutData[OutLength++]=InData[Pointer++];
 }
 free(HashTable4);
 free(HashTable5);
 return OutLength;
}

uint LZ_UnPreprocess(uc *InData, uc *OutData, uint InLength)
{
  uint Pointer,i,CommonLength,OutLength,HashIndex4,HashIndex5;
  uc *Pointer4, *Pointer5, *PastPointer;

 HashTable4=(uc **)malloc(HTSIZE4*(uint)sizeof(uc *));
 HashTable5=(uc **)malloc(HTSIZE5*(uint)sizeof(uc *));

 CleanTabs();

 OutLength=0;
 for (i=0; i<5; i++) OutData[i]=InData[i];
 OutLength+=5;

 Pointer=5;

 HashTable4[HashFunction4(4,OutData)]=OutData+4;
 while (Pointer<InLength) {

       HashIndex4=HashFunction4(OutLength,OutData);
       HashIndex5=HashFunction5(OutLength,OutData);

       Pointer4=HashTable4[HashIndex4];
       Pointer5=HashTable5[HashIndex5];

       HashTable5[HashIndex5]=HashTable4[HashIndex4]=OutData+OutLength;

       if (Pointer5!=NULL || Pointer4!=NULL) {
          if (Pointer5!=NULL) PastPointer=Pointer5;
           else PastPointer=Pointer4;
          CommonLength=GetLength(InData,&Pointer);
          if (CommonLength<256)
             OutData[OutLength++]=(uc)CommonLength;
           else {
             CommonLength=CommonLength-256+LowerLimit;
             while (CommonLength--)
                   OutData[OutLength++]=*PastPointer++;
           }
       } else  OutData[OutLength++]=InData[Pointer++];
 }
 free(HashTable4);
 free(HashTable5);
 return OutLength;
}
