#define ST_MaxByte           256
#define ST_MaxWord           65536
#define ST_INDIRECT          0x800000

sint GRZip_ST4_Encode(uc * Input,sint Size)
{
  sint *  Counter=(sint *)malloc(ST_MaxWord*sizeof(sint));
  uint *  Context=(uint *)malloc(Size*sizeof(uint));
  sint    FBP,i;
  uint    W;
  memset    (Counter,0,ST_MaxWord*sizeof(sint));

  W=Input[Size-1]<<8; for (i=0;i<Size;i++) Counter[(W=(W>>8)|(Input[i]<<8))]++;

  for (W=0,i=0;i<ST_MaxWord;i++) W+=Counter[i],Counter[i]=W-Counter[i];

  W=(Input[Size-4]<<8)|Input[Size-5];
  if (W==0xFFFF) FBP=Size-1; else FBP=Counter[W+1]-1;

  W=(Input[Size-1]<<24)|(Input[Size-2]<<16)|(Input[Size-3]<<8)|(Input[Size-4]);
  
  for (i=0;i<Size;i++)
  {
    uc c=Input[i];
    Context[Counter[W&0x0000FFFF]++]=(W&0xFFFF0000)|c;
    W=(W>>8)|(c<<24);
  }

  for (i=Size-1;i>=FBP;i--) Input[--Counter[Context[i]>>16]]=Context[i]&0xFF;
  FBP=Counter[Context[FBP]>>16];
  for (;i>=0;i--) Input[--Counter[Context[i]>>16]]=Context[i]&0xFF;
  
  free(Context);
  free(Counter);
  return FBP;
}

#define ST_SetBit(Bit) (Flag[Bit>>3]|=(1<<(Bit & 7)))
#define ST_GetBit(Bit) (Flag[Bit>>3]&(1<<(Bit & 7)))

void GRZip_ST4_Decode(uc * Input,sint Size,sint FBP)
{
  sint *  Context2=(sint *)malloc(ST_MaxWord*sizeof(sint));
  uc   *  Flag=(uc *)malloc(((Size+8)>>3)*sizeof(uc));
  uint *  Table=(uint *)malloc((Size+1)*sizeof(uint));
  sint    LastSeen[ST_MaxByte],T[ST_MaxByte],S[ST_MaxByte];
  sint    CStart,Sum,i,j;

  memset    (Context2,0,ST_MaxWord*sizeof(sint));
  memset    (Flag,0,((Size+8)>>3)*sizeof(uc));
  
  memset    (T,0,ST_MaxByte*sizeof(sint));
  memset    (LastSeen,0xFF,ST_MaxByte*sizeof(sint));
  
  for (i=0;i<Size;i++) T[Input[i]]++;

  for (i=0,j=0,Sum=0;i<ST_MaxByte;i++)
    for (Sum+=T[i],T[i]=Sum-T[i];j<Sum;j++)
      Context2[(Input[j]<<8)|i]++;

  memcpy(S,T,ST_MaxByte*sizeof(sint));

  for (i=0,j=0,Sum=0;i<ST_MaxWord;i++)
    for (CStart=Sum,Sum+=Context2[i];j<Sum;j++)
    {
      uc c=Input[j];
      if (LastSeen[c]!=CStart)
      {
        LastSeen[c]=CStart;
        ST_SetBit(T[c]);
      }
      T[c]++;
    }

  memset(LastSeen,0,ST_MaxByte*sizeof(sint));
  
  for (CStart=0,i=0;i<Size;i++)
  {
    uc c=Input[i];
    if (ST_GetBit(i)) CStart=i;
    if (LastSeen[c]<=CStart)
    {
      Table[i]=S[c];
      LastSeen[c]=i+1;
    }
    else
      Table[i]=(LastSeen[c]-1)|ST_INDIRECT;
    S[c]++;
    Table[i]|=(c<<24);
  }
  Table[Size]=ST_INDIRECT;

  free(Context2);
  free(Flag);

  for (j=FBP,Sum=Table[FBP],i=0;i<Size;i++)
    if (Sum&ST_INDIRECT)
    {
      j=((Table[Sum&(ST_INDIRECT-1)]++)&(ST_INDIRECT-1));
      Sum=Table[j];
      Input[i]=Sum>>24;
    }
    else
    {
      Table[j]++; Sum=Table[j=Sum&(ST_INDIRECT-1)];
      Input[i]=Sum>>24;      
    }

  free(Table);
}

#undef ST_MaxByte
#undef ST_MaxWord
#undef ST_SetBit
#undef ST_GetBit
